import argparse
import functools
import inspect


class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__subparsers = None

    @property
    def subparsers(self):
        if self.__subparsers is None:
            self.__subparsers = self.add_subparsers()
        return self.__subparsers

    def subcommand(self, func=None, **subcommand_options):
        """
        Decorator to register a function as a subcommand.

        Example:

            @parser.subcommand(name="list")
            def list_cmd(args):
                pass

            @parser.subcommand(
                name="describe",
                args=[
                    ["item_id", {"help": "Id of item to describe}]
                ],
            )
            def describe_cmd(args):
                pass

        """

        def decorator(subcommand_func):
            subcommand_sig = inspect.signature(subcommand_func)

            @functools.wraps(subcommand_func)
            def wrapped(args):

                final_args = []
                final_kwargs = {}

                if "args" in subcommand_sig.parameters:
                    final_kwargs["args"] = args

                return subcommand_func(*final_args, **final_kwargs)

            subcommand_name = subcommand_options.pop("name", subcommand_func.__name__)
            subcommand_args_def = subcommand_options.pop("args", None) or ()
            subcommand_doc = subcommand_options.pop("help", None) or subcommand_options.pop("description", None)
            if subcommand_doc is None:
                subcommand_doc = subcommand_func.__doc__
            subcommand_aliases = subcommand_options.pop("aliases", None) or []
            if subcommand_options:
                raise ValueError(f"Unexpected kwarg(s): {', '.join(str(k) for k in subcommand_options.keys())}")

            parser = self.subparsers.add_parser(
                name=subcommand_name,
                help=subcommand_doc,
                description=subcommand_doc,
                aliases=subcommand_aliases,
            )
            parser.set_defaults(func=wrapped)

            for arg in subcommand_args_def:
                if isinstance(arg, str):
                    parser.add_argument(arg)
                elif isinstance(arg, (list, tuple)):
                    if isinstance(arg[-1], dict):
                        arg_args = arg[:-1]
                        arg_kwargs = arg[-1]
                    else:
                        arg_args = arg[:]
                        arg_kwargs = {}
                    parser.add_argument(*arg_args, **arg_kwargs)
                else:
                    raise TypeError(r"Expected a string or list, got {arg!r}")

            return parser

        if func is None:
            return decorator
        else:
            return decorator(func)



import functools
from typing import Callable, Type

from .func_utils import optional_args_func
from .parsers import ArgumentParser


def cli(init_parser: Callable) -> Type["Cli"]:
    """
    Decorator for a function that takes "parser: ArgumentParser" as argument and initialises it as required.
    """

    cli_help = init_parser.__doc__
    init_parser = optional_args_func(init_parser)

    class CustomCli(Cli):
        def __init__(self):
            super().__init__()
            self.parser.description = cli_help
            init_parser(parser=self.parser, subcommand=self.parser.subcommand)

    CustomCli.__name__ = init_parser.__name__
    CustomCli.__qualname__ = init_parser.__qualname__

    return CustomCli


class _WithInstance:
    """
    This allows the method to be called with a class as well as with instance.
    If called with a class it will create a new instance and return a method
    bound to that instance.
    """

    def __init__(self, method):
        self.method = method

    def __set_name__(self, owner, name):
        self.__name__ = name

    def __get__(self, instance, owner):
        if instance is None:
            instance = owner()
        return functools.partial(self.method, instance)


class Cli:
    parser: ArgumentParser

    def __init__(self):
        self.parser = ArgumentParser()
        self.parser.set_defaults(func=optional_args_func(self.parser.print_help))

    def _run(self, args=None):
        parsed_args = self.parser.parse_args(args=args)
        parsed_args.func(args=parsed_args)

    run: Callable = _WithInstance(_run)



import functools
import inspect


def optional_args_func(func) -> callable:
    """
    Given a function or generator `func`, return a function/generator
    that takes any number of kwargs and calls `func` with only the args/kwargs
    that `func` expects.
    """
    if getattr(func, "_optional_args_func", False):
        return func

    is_generator = inspect.isgeneratorfunction(func)
    func_sig = inspect.signature(func)
    expects_nothing = not func_sig.parameters

    if is_generator:
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            if expects_nothing:
                yield from func()
            else:
                bound_arguments = func_sig.bind(*args, **{k: v for k, v in kwargs.items() if k in func_sig.parameters})
                yield from func(*bound_arguments.args, **bound_arguments.kwargs)
    else:
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            if expects_nothing:
                return func()
            else:
                bound_arguments = func_sig.bind(*args, **{k: v for k, v in kwargs.items() if k in func_sig.parameters})
                return func(*bound_arguments.args, **bound_arguments.kwargs)

    # Mark it so that we don"t double wrap our own
    setattr(wrapped, "_optional_args_func", True)

    return wrapped



import argparse
import inspect


def arg_converter(func=None, **options):
    """
    Decorator to declare simple argparse argument value converters without argparse.Action boilerplate

    Example:

        @arg_converter
        def int_value(value):
            return int(value)

        parser.add_argument('x', action=int_value)

    """

    def decorator(converter_func):
        sig = inspect.signature(converter_func)

        def action_func(action, parser, args, value, option_string=None):
            all_args = {
                'action': action,
                'parser': parser,
                'args': args,
                'value': value,
                'option_string': option_string,
            }
            converter_args = {k: all_args[k] for k in sig.parameters}
            setattr(args, action.dest, converter_func(**converter_args))

        if hasattr(func, "__name__"):
            name = getattr(func, "__name__")
        else:
            name = func.__class__.__name__

        return type(
            '{}#argparse.Action'.format(name),
            (argparse.Action,),
            {'__call__': action_func}
        )

    if func is None:
        return decorator
    else:
        return decorator(func)



__version__ = "0.1.0"  # also in "setup.py"


from .arguments import arg_converter
from .cli import cli
from .parsers import ArgumentParser

__all__ = [
    "arg_converter",
    "cli",
    "ArgumentParser",
]
